/*
 * Copyright (C) 1996-8 Michael R. Elkins <me@cs.hmc.edu>
 * Copyright (C) 1996-9 Brandon Long <blong@fiction.net>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 


#include "mutt.h"
#include "mutt_curses.h"
#include "mx.h"
#include "mailbox.h"
#include "globals.h"
#include "sort.h"
#include "smtp.h"

#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <stdio.h>


char *smtp_reply_error[] = 
{
  NULL, 
  "ESTABILISHED",
  "HELO",
  "EHLO", 
  "MAIL",
  "RCPT",
  "DATA",
  "RSET",
  "NOOP",
  "QUIT"
};

int smtp_parse_reply ( char *buf )
{
  enum { SMTPSUCCESS = 1 ,SMTPINTERMEDIATE ,SMTPFAILURE ,SMTPERROR };
  int count1 ,count2;
   
  char *code[][6][10] = 
     { { {NULL },
	 {NULL },                           /*SUCCESS*/
	 {NULL },                           /*INTERMEDIATE*/      
	 {NULL },                           /*FAILURE*/
	 {NULL },                           /*ERROR*/
         {NULL ,NULL }
       },
	
       { {NULL},
	 {NULL,"220",NULL},                                    /*ESTABILISHED*/
	 {NULL,"666",NULL},
	 {NULL,"421",NULL},
	 {NULL,NULL },
	 {NULL,NULL }
       },

       { {NULL},
	 {NULL,"250",NULL},                                    /*HELO*/
	 {NULL,"666",NULL},
	 {NULL,"666",NULL},
	 {NULL,"500","501","504","421",NULL},
         {NULL,NULL }
       },
	
       { {NULL},
	 {NULL,"250",NULL},                                      /*EHLO*/
	 {NULL,"666",NULL},
	 {NULL,"666",NULL},
	 {NULL,"500",NULL},
         {NULL,NULL }
       },
	
       { {NULL},
	 {NULL,"250",NULL},                                      /*MAIL*/
	 {NULL,"666",NULL},
	 {NULL,"552","451","452",NULL},
	 {NULL,"500","501","421",NULL},
         {NULL,NULL}
       },
	
       { {NULL},
	 {NULL,"250","251",NULL},                                /*RCPT*/
	 {NULL,"666",NULL },
	 {NULL,"550","551","552","553","450","451","452",NULL},
	 {NULL,"500","501","503","421",NULL},
	 {NULL,NULL}
       },
	
       { {NULL},
	 {NULL,"250","354",NULL},                                /*DATA*/
	 {NULL,"354",NULL},
	 {NULL,"552","554","451","452",NULL},
	 {NULL,"500","501","503","421",NULL},
         {NULL,NULL }
       },
	
       { {NULL},
	 {NULL,"250",NULL},                                     /*RSET*/
	 {NULL,"666",NULL},
	 {NULL,"666",NULL},
	 {NULL,"500","501","504","421",NULL},
         {NULL,NULL}
       },
	
       { {NULL},
	 {NULL,"250",NULL},                                     /*NOOP*/ 
	 {NULL,"666",NULL},
	 {NULL,"666",NULL},
	 {NULL,"500","421",NULL},
         {NULL,NULL}
       },
	
       { {NULL},
	 {NULL,"221",NULL},                                     /*QUIT*/
	 {NULL,"666",NULL},
	 {NULL,"666",NULL},
	 {NULL,"500",NULL},
         {NULL,NULL}
       },
	
       { {NULL},
	 {NULL ,NULL}
       }
	  
     };


  count1 = 1; 

  /*
  mutt_message (_("Receiving \"%s\""), buf);
  sleep(1);
  */

  while ( code[smtp_data.command][count1][1] ) {
      
      count2 =1 ;
      
      while (code[smtp_data.command][count1][count2]) {
	 
	    if (!strncmp (buf,code[smtp_data.command][count1][count2],3) )
	       goto end;
	    count2++;
      }
      
      count1++;
  } 
   
   
   
  if (!code[smtp_data.command][count1][1]) {

     if ( (smtp_data.command == SMTP_COMMAND_RSET  ||
	   smtp_data.command == SMTP_COMMAND_QUIT) && 
	  (smtp_data.status == SMTP_REPLY_ERROR    ||
	   smtp_data.status == SMTP_REPLY_UNKNOW_ERROR )  ) return (-1) ;

     
     smtp_data.status = SMTP_REPLY_UNKNOW_ERROR;
     
     snprintf (smtp_data.reply, sizeof (smtp_data.reply), 
	       "REPLY UNKNOW (%s) %s \r\n",
	       smtp_reply_error[smtp_data.command] ,buf );
     
     return (-1);
     
  }

end:   
   
  if ( count1 == SMTPSUCCESS ) return (0); 
   
  else {
     
       if ( (smtp_data.command == SMTP_COMMAND_RSET  ||
	     smtp_data.command == SMTP_COMMAND_QUIT) && 
	    (smtp_data.status == SMTP_REPLY_ERROR    ||
	     smtp_data.status == SMTP_REPLY_UNKNOW_ERROR )  ) return (-1) ;
       
	  
     
       smtp_data.status = SMTP_REPLY_ERROR ;
     
       snprintf (smtp_data.reply, sizeof (smtp_data.reply), 
		 "REPLY ERROR (%s) %s \r\n",
		 smtp_reply_error[smtp_data.command] ,buf );
     
       return (-1);
  }
   
}

static int getLine (int fd, char *s, int len)
{
  char ch;
  int bytes = 0;

  while (read (fd, &ch, 1) > 0)
  {
    *s++ = ch;
    bytes++;
    if (ch == '\n')
    {
      *s = 0;
      return (bytes);
    }
    /* make sure not to overwrite the buffer */
    if (bytes == len - 1)
    {
      *s = 0;
      return bytes;
    }
  }
  *s = 0;
  return (-1);
}

static int send_smtp(int fd, char *s, int len)
{
  /*
  mutt_message (_("Sending \"%s\""), s);
  sleep(1);
  */
  return write (fd,s,len);
}

static int smtp_send_rcpt (ADDRESS *rcpt)
{
  char buf[LONG_STRING] ,tmpbuf[LONG_STRING];

  while (rcpt)
  {
    snprintf (tmpbuf, sizeof (tmpbuf), "RCPT TO:<%s>\r\n",rcpt->mailbox);   
   
    if ( send_smtp ( smtp_data.fd ,tmpbuf, mutt_strlen(tmpbuf) ) < 0 )
    {   
      close (smtp_data.fd);
      smtp_data.state  = SMTP_CONNECTED;
      smtp_data.status = SMTP_SOCKET_WRITE_ERROR;
      snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf); 
      return (-1);
    } 
   
    if (getLine (smtp_data.fd, buf, sizeof (buf)) < 0)
    {
      close (smtp_data.fd);
      smtp_data.state  = SMTP_CONNECTED;
      smtp_data.status = SMTP_SOCKET_READ_ERROR;
      snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,buf);
      return (-1);
    }     

    smtp_data.command = SMTP_COMMAND_RCPT ;  
   
    if ( smtp_parse_reply ( buf) ) { return -2 ; }

    rcpt = rcpt->next;
  }
  return 0;
}

int smtp_send_message (ADDRESS *from,	/* the sender */
		 ADDRESS *to, ADDRESS *cc, ADDRESS *bcc, /* recips */
		 const char *message) /* file containing message */
{
  int  Flush = 10000 ,count ,total ,dummy, tmpc=0 ;
  char buf[LONG_STRING] ,tmpbuf[LONG_STRING] ,tmpf[10003];
  char host[SHORT_STRING] ,user[SHORT_STRING] ,serv[SHORT_STRING];
  char *tmpff ;
  char *slang[] = { "|" ,"/" ,"-" ,"\\" ,NULL };
  FILE *in=NULL ;

  struct sockaddr_in sin;
#if SIZEOF_LONG == 4
  long n;
#else
  int n;
#endif
  struct hostent *he;

  smtp_data.status = smtp_data.state = 0 ; 
   
  if (!SmtpPort ) SmtpPort = 25 ;

  if (!SmtpServ) 
  { 
     serv[0] = 0 ;
     
     if (mutt_get_field(_("SMTP Server: "),serv ,sizeof(serv),0)!=0 ||
	 !serv[0] ) 
     {
      serv[0] = 0 ;
      smtp_data.state  = SMTP_DISCONNECTED;
      smtp_data.status = SMTP_ENTER_DATA_ERROR;
      return (-1);
     }
  
  }
  else strfcpy ( serv ,SmtpServ ,sizeof (serv) );	   

  if (!SmtpUser && !(option (OPTENVFROM) && from && !from->next))
  { 
     user[0] = 0 ;
     
     if (mutt_get_field(_("SMTP User: "),user ,sizeof(user),0)!=0 ||
	 !user[0] ) 
     {
      user[0] = 0 ;
      smtp_data.state  = SMTP_DISCONNECTED;
      smtp_data.status = SMTP_ENTER_DATA_ERROR;
      return (-1);
     }
  
  }
  else if (SmtpUser)
    strfcpy ( user ,SmtpUser ,sizeof (user) );	   

  if (!SmtpHost && !Fqdn) 
  { 
     host[0] = 0 ;

     if (mutt_get_field(_("SMTP Host: "),host ,sizeof(host),0)!=0 ||
	 !host[0] ) 
     {
      host[0] = 0 ;
      smtp_data.state  = SMTP_DISCONNECTED;
      smtp_data.status = SMTP_ENTER_DATA_ERROR;
      return (-1);
     }
  
  }
  else strfcpy ( host ,SmtpHost ? SmtpHost:Fqdn ,sizeof (host) );	   
   
/*   
  if (!SmtpPass) 
  { 
     pass[0] = 0 ;
     
     snprintf (buf ,sizeof(buf) ,_("password for %s@%s: "),user ,conn->server);

     if (mutt_get_field (buf,pass ,sizeof(pass),M_PASS) != 0 || !pass[0] ) 
     {
      smtp_data.state  = SMTP_DISCONNECTED;
      smtp_data.status = SMTP_ENTER_DATA_ERROR;	
      return (-1);
     }
  }
  else strfcpy (pass ,SmtpPass ,sizeof (pass));	
*/

  smtp_data.fd = socket (AF_INET, SOCK_STREAM, IPPROTO_IP);
   
  memset ((char *) &sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_port = htons (SmtpPort);

  if ((n = inet_addr (NONULL(serv))) == -1)
  {
    /* Must be a DNS name */
    if ((he = gethostbyname (NONULL(serv))) == NULL)
    {
      mutt_error (_("Could not find address for host %s."), serv);
      smtp_data.state  = SMTP_DISCONNECTED;
      smtp_data.status = SMTP_CONNECTION_ERROR; 
      return -1;
    }
    memcpy ((void *)&sin.sin_addr, *(he->h_addr_list), he->h_length);
  }
  else
    memcpy ((void *)&sin.sin_addr, (void *)&n, sizeof(n));
  
  mutt_message (_("Connecting to %s"), inet_ntoa (sin.sin_addr));

  if (connect (smtp_data.fd, (struct sockaddr *) &sin, sizeof (struct sockaddr_in)) == -1)
  {
    mutt_perror ("connect");
    smtp_data.state  = SMTP_DISCONNECTED;
    smtp_data.status = SMTP_CONNECTION_ERROR; 
    return -1;
  }

  smtp_data.state = SMTP_CONNECTED;

  if (getLine (smtp_data.fd, buf, sizeof (buf)) < 0)
  {
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_READ_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,buf);
    return (-1);
  } 

  smtp_data.command = SMTP_ESTABILISHED ; 

  if ( smtp_parse_reply ( buf) ) {close (smtp_data.fd) ;return (-1) ; }
   
/**** HELO **********************************/

  snprintf (tmpbuf, sizeof (tmpbuf), "HELO %s\r\n", host );
   
  if ( send_smtp ( smtp_data.fd ,tmpbuf, mutt_strlen(tmpbuf) ) < 0 )
  {    
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_WRITE_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf); 
    return (-1);
  }   
   
  if (getLine (smtp_data.fd, buf, sizeof (buf)) < 0)
  {
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_READ_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,buf);
    return (-1);
  }     

  smtp_data.command = SMTP_COMMAND_HELO ;  
   
  if ( smtp_parse_reply ( buf) ) { goto QUITSMTP ; }   
   
   
/****  MAIL *********************************/   

  if (option (OPTENVFROM) && from && !from->next) {
    snprintf(tmpbuf,sizeof (tmpbuf),"MAIL FROM:<%s>\r\n" ,from->mailbox);
  }
  else
  {
    snprintf(tmpbuf,sizeof (tmpbuf),"MAIL FROM:<%s@%s>\r\n" ,user,host);
  }

  if ( send_smtp ( smtp_data.fd ,tmpbuf, mutt_strlen(tmpbuf) ) < 0 )
  {   
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_WRITE_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf); 
    return (-1);
  } 

  if (getLine (smtp_data.fd, buf, sizeof (buf)) < 0)
  {
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_READ_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,buf);
    return (-1);
  }     

  smtp_data.command = SMTP_COMMAND_MAIL ;  
   
  if ( smtp_parse_reply ( buf) ) { goto  RESETSMTP ; }

/***  RCPT ***********************************/   

  
  dummy = smtp_send_rcpt (to);
  if (!dummy) dummy = smtp_send_rcpt (cc);
  if (!dummy) dummy = smtp_send_rcpt (bcc);

  if ( dummy == -1 ) return -1;
  if ( dummy == -2 ) goto RESETSMTP;

/****  DATA ***********************************/
  
  snprintf (tmpbuf, sizeof (tmpbuf), "DATA\r\n");   
   
  if ( send_smtp ( smtp_data.fd ,tmpbuf, mutt_strlen(tmpbuf) ) < 0 )
  {   
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_WRITE_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf); 
    return (-1);
  } 
   
  if (getLine (smtp_data.fd, buf, sizeof (buf)) < 0)
  {
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_READ_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,buf);
    return (-1);
  }     

  smtp_data.command = SMTP_COMMAND_DATA ;  
   
  if ( smtp_parse_reply ( buf) ) { goto RESETSMTP ; }
   
/***** message ********************************/

  if ( (in = fopen ( message ,"rb" ) ) == NULL  ){
     mutt_error _("Can't Open Message ");
     smtp_data.status = SMTP_FATAL ;
     return (-1);
  }
   
  total = 0 ;dummy = 0; 

ciclo:   
   
  count = 0 ;tmpff = tmpf ; 
   
  while ( (count < Flush ) && (tmpc = getc (in) ) != EOF ) {
       
        *tmpff = tmpc;
        count++ ;tmpff++ ;

	if (count==1 && tmpc=='.') {
	    *(tmpff++) = '.';
	    count++;
	}

        if ( (*(tmpff-1) == '\n') ) {
	  if ( (*(tmpff-2) != '\r') ) {
	    *(tmpff-1) = '\r';
	    *(tmpff++) = '\n';
	    count++;
	  }
	  break;
	}
  }

  total += count ; 
   
  if ( (*(tmpff-1) != '\n') ) {
    *(tmpff++) = '\r' ;
    *(tmpff++) = '\n' ;
  }
  *tmpff = '\0' ;
   
  if ( count && write ( smtp_data.fd ,tmpf, mutt_strlen(tmpf) ) < 0 )
  {   
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_WRITE_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf); 
    if (in) fclose (in);
    return (-1);
  } 
   
  mutt_message( _("Sending... %c  Sent %d" ) ,*slang[dummy],total );

  dummy++ ; 
   
  dummy %= 4 ; 
   
  if ( tmpc != EOF ) goto ciclo; 
   
   
  snprintf (tmpbuf, sizeof (tmpbuf), ".\r\n" );   
   
  if ( send_smtp ( smtp_data.fd ,tmpbuf, mutt_strlen(tmpbuf) ) < 0 )
  {   
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_WRITE_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf); 
    if (in) fclose (in);
    return (-1);
  } 

  if (getLine (smtp_data.fd, buf, sizeof (buf)) < 0)
  {
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_READ_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,buf);
    if (in) fclose (in);
    return (-1);
  }     

  smtp_data.command = SMTP_COMMAND_DATA ;  
   
  if ( smtp_parse_reply ( buf) ) { goto RESETSMTP; }

/*****  QUIT  ************************************/

QUITSMTP:   
  
  snprintf (tmpbuf, sizeof (tmpbuf), "QUIT\r\n");   
   
  if ( send_smtp ( smtp_data.fd ,tmpbuf, mutt_strlen(tmpbuf) ) < 0 )
  {   
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_WRITE_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf); 
    if (in) fclose (in);
    return (-1);
  } 

  if (getLine (smtp_data.fd, tmpbuf, sizeof (tmpbuf)) < 0)
  {
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_READ_ERROR;
    if (tmpbuf[0])
      snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf);
    else
      snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,buf);
    if (in) fclose (in);
    return (-1);
  }     

  smtp_data.command = SMTP_COMMAND_QUIT ;  
   
  if ( smtp_parse_reply ( tmpbuf) ) {
    close (smtp_data.fd) ;
    if (in) fclose (in);
    return (-1) ; 
  }

  if ( smtp_data.status == SMTP_REPLY_ERROR ||
       smtp_data.status == SMTP_REPLY_UNKNOW_ERROR ) {
    if (in) fclose (in);
    return (-1) ;
  }
   
  else {  
     close (smtp_data.fd) ;
     if (in) fclose (in);
     return (0);
  }

/******* RSET **********************************/   
   
RESETSMTP:
   
  snprintf (tmpbuf, sizeof (tmpbuf), "RSET\r\n");   
   
  if ( send_smtp ( smtp_data.fd ,tmpbuf, mutt_strlen(tmpbuf) ) < 0 )
  {   
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_WRITE_ERROR;
    snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf); 
    if (in) fclose (in);
    return (-1);
  } 

  if (getLine (smtp_data.fd, tmpbuf, sizeof (tmpbuf)) < 0)
  {
    close (smtp_data.fd);
    smtp_data.state  = SMTP_CONNECTED;
    smtp_data.status = SMTP_SOCKET_READ_ERROR;
    if (tmpbuf[0])
      snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,tmpbuf);
    else
      snprintf ( smtp_data.error ,sizeof (smtp_data.error) ,buf);
    if (in) fclose (in);
    return (-1);
  }     

  smtp_data.command = SMTP_COMMAND_RSET ;  
   
  if ( smtp_parse_reply ( tmpbuf) ) {
    close (smtp_data.fd) ;
    if (in) fclose (in);
    return (-1) ; 
  }
 
  else goto QUITSMTP;
  
}



int mutt_invoke_smtp (ADDRESS *from,	/* the sender */
		 ADDRESS *to, ADDRESS *cc, ADDRESS *bcc, /* recips */
		 const char *msg) /* file containing message */
{
  int i;

  smtp_data.state = 0;
  smtp_data.status = 0;
  smtp_data.error[0] = '\0';
  smtp_data.reply[0] = '\0';

  i = smtp_send_message (from, to,cc,bcc, msg);

  unlink (msg);

  if ( i == -1 ) {
    if (smtp_data.status == SMTP_ENTER_DATA_ERROR ) 
      mutt_error _("Data Error") ;
    if (smtp_data.status == SMTP_CONNECTION_ERROR ) 
      mutt_error _("Connection Error") ;
    if (smtp_data.status == SMTP_SOCKET_READ_ERROR ) 
      mutt_error ( _("Read %s"),smtp_data.error) ;
    if (smtp_data.status == SMTP_SOCKET_WRITE_ERROR ) 
      mutt_error ( _("Write %s"),smtp_data.error) ;
    if (smtp_data.status == SMTP_REPLY_ERROR ||
        smtp_data.status == SMTP_REPLY_UNKNOW_ERROR ) 
      mutt_error ( _("%s"),smtp_data.reply);
  }
  else if ( i == 0 ) {
    mutt_error ( _("OK !!! %s"),smtp_data.reply );
    return (0);
  }

  return i;
}
